import torch.nn as nn
import torch.nn.functional as F
from .args import args
from . import mpolinear

class Net(nn.Module):
    def __init__(self,args):
        super(Net, self).__init__()
        self.args = args
        D=args.D
        print("bias=",args.bias)
        dims = [int(i) for i in args.dims.split(",")]
        assert( len(dims) == 4)
        if(args.net == 'fc'):
            print("FC layer, D=",args.D)
            self.fc1 = nn.Linear(784, D,bias=args.bias)
            if args.bn:
                self.bn = nn.BatchNorm1d(D)
            self.fc2 = nn.Linear(D, 10,bias=args.bias)
        elif(args.net == 'lrfc'): # low rank fully connected
            print("Low rank FC layer, chi=",args.chi)
            self.mpofc1 = mpolinear.LRLinear(784, D, bias=args.bias,chi=args.chi)
            self.fc2 = nn.Linear(D, 10)
        elif(args.net == 'mpofc'):
            print("MPOFC2 layer, chi=",args.chi)
            self.mpofc1 = mpolinear.MPOLinear2([4,7,7,4],dims, bias=args.bias,chi=args.chi)
            self.mpofc2 = mpolinear.MPOLinear2(dims,[1,1,10,1], bias=args.bias,chi=4)
        elif(args.net == 'mpofc3'):
            print("MPOFC3 layer, chi=",args.chi)
            self.mpofc1 = mpolinear.MPOLinear3([4,7,7,4],dims, bias=args.bias,chi=args.chi)
            self.mpofc2 = mpolinear.MPOLinear3(dims,[1,1,10,1], bias=args.bias,chi=4)
        else:
            raise ValueError('Unknown net: {}'.format(args.net))

    def forward(self, x):
        x = x.view(x.shape[0],-1)
        if(self.args.net == 'fc'):
            x = F.relu(self.fc1(x))
            if self.args.bn:
                x = self.bn(x)
            return F.log_softmax(self.fc2(x), dim=1)
        else:
            re = F.relu(self.mpofc1(x))
            return F.log_softmax(self.mpofc2(re), dim=1)
