import torch.nn as nn
import numpy as np
import torch.nn.functional as F
from .args import args
from . import mpolinear
from .utils import AttrProxy

class Net(nn.Module):
    def __init__(self,args,indim = 784):
        super(Net, self).__init__()
        D=args.D
        self.args = args
        print("bias=",args.bias)
        self.dims = []
        self.indim = indim
        for dim in args.dims.split(';'):
            self.dims.append([int(i) for i in dim.split(",")])
        if self.indim == 784 and np.prod(self.dims[0]) != 784:
            self.dims.insert(0,[4,7,7,4])
        print("input dimension=",self.indim,"MPO indim=",np.prod(self.dims[0]),)
        assert(np.prod(self.dims[0]) == self.indim)
        if np.prod(self.dims[-1]) != 10:
            self.dims.append([1,1,10,1])
        print("dims:",self.dims)

        for i in range(len(self.dims)-1):
            exec("self.layers_"+str(i)+"=mpolinear.OperatorLinear(self.dims[i],self.dims[i+1], bias=args.bias,chi=args.chi)")
            if args.bn and i<len(self.dims)-1:
                exec("self.bns_"+str(i)+"=nn.BatchNorm1d(np.prod(self.dims[i+1]))")
        self.layers = AttrProxy(self, 'layers_')
        self.bns = AttrProxy(self, 'bns_')

    def forward(self, x):
        x = x.view(x.shape[0],-1)
        for i in range(len(self.dims)-2):
            x = F.relu(self.layers[i](x))
            if self.args.bn:
                x = self.bns[i](x)
        return F.log_softmax(self.layers[len(self.dims)-2](x), dim=1)
