import math
import torch.nn as nn
import numpy as np

import torch
from torch.nn.parameter import Parameter
from torch.nn import functional as F
from torch.nn import init
from torch.nn.modules.module import Module
#from torch._jit_internal import weak_module, weak_script_method
#from TorNcon import ncon


#@weak_module
class PEPO(Module):
    def __init__(self, L, d,D=2):
        """
        L*L lattice, channel d
        """
        super(PEPO, self).__init__()
        self.d = d
        self.L = L
        self.D=D
        print("Constructing PEPO, d=",d,"%d x %d"%(L,L))
        scale=0.1
        self.tensors = Parameter(torch.randn(L,L,d,10,D,D,D,D))
        self.beta=Parameter(torch.tensor([100.0]))

        print(self)
        print("Parameters in the class")
        params = list(self.parameters())
        params = list(filter(lambda p: p.requires_grad, params))
        nparams = int(sum([np.prod(p.shape) for p in params]))
        print('Total number of trainable parameters: {}'.format(nparams))
        for param in self.parameters():
            print(type(param.data), param.size())
        print("done")
        print(self.tensors.shape)


#    @weak_script_method
    def forward2(self, x_input):
        #        print("x_input shape",x_input.shape)
#        print("intput",x_input)
        tensors=torch.einsum("abcd,cdbeijkl->acdeijkl",x_input,self.tensors) # batch,l,l,10,left,down,right,up
        batch = tensors.shape[0]
        D=self.D
        L=self.L
#        print("tensors.shape=",tensors.shape)
        mps_top = torch.zeros(batch,L,10,D,D,D,dtype=x_input.dtype,device=x_input.device) # batch,l,10,left,down,right
        mps_top = tensors[:,0,:,:,:,:,:,0]# batch,l,10,left,down,right
#        print("mps_top",mps_top)
        mps_bottom = torch.zeros(batch,L,10,D,D,D,dtype=x_input.dtype,device=x_input.device) # batch,l,10,left,up,right
        mps_bottom = tensors[:,L-1,:,:,:,0,:,:]# batch,l,10,left,down,right
        for i in range(1,L>>1):
            mps_top_new = torch.zeros(batch,L,10,mps_top.shape[3]*D,D,mps_top.shape[5]*D,dtype=x_input.dtype,device=x_input.device) # batch,l,10,left,down,right
            mps_bottom_new = torch.zeros(batch,L,10,mps_bottom.shape[3]*D,D,mps_bottom.shape[5]*D,dtype=x_input.dtype,device=x_input.device) # batch,l,10,left,down,right
#            print("eat row",i)
            for j in range(L):
                #                print(j,'th tensor')
                mps_top_new[:,j,:,:,:,:] = torch.einsum("xyabc,xyijkb->xyaijck",mps_top[:,j,:,:,:,:],tensors[:,i,j,:,:,:,:,:]).contiguous().view(batch,10,mps_top.shape[3]*D,D,mps_top.shape[5]*D) # x:batch y:10 
                mps_bottom_new[:,j,:,:,:,:] = torch.einsum("xyabc,xyibjk->xyiakjc",mps_bottom[:,j,:,:,:,:],tensors[:,L-1-i,j,:,:,:,:,:]).contiguous().view(batch,10,mps_bottom.shape[3]*D,D,mps_bottom.shape[5]*D) # x:batch y:10 
            print(mps_top_new)
            input("mps_top_new.norm %.3e"%mps_top_new.norm())
            print(mps_bottom_new)
            input("mps_bottom_new  %.3e"%mps_bottom_new.norm())
#            mps_top_new = mps_top_new / (mps_top_new.norm()+1.0e-20)
#            norms = mps_top_new * mps_top_new
#            mps_bottom_new = mps_bottom_new / (mps_bottom_new.norm()+1.0e-20)
            mps_top = mps_top_new
            mps_bottom = mps_bottom_new
#        print("mps_top:",mps_top[:10,0,0,0,0,0])
#        print("mps_bottom:",mps_bottom[:10,0,0,0,0,0])
        lmat = torch.einsum("xyab,xyac->xybc",mps_top[:,0,:,0,:,:],mps_bottom[:,0,:,0,:,:])
        rmat = torch.einsum("xyab,xycb->xyac",mps_top[:,0,:,:,:,0],mps_bottom[:,0,:,:,:,0])
        for j in range(1,L>>1):
            lmat = torch.einsum("xyab,xyacd,xybcf->xydf",lmat,mps_top[:,j,:,:,:,:],mps_bottom[:,j,:,:,:,:]).contiguous().view(batch,10,mps_top.shape[5],mps_bottom.shape[5])
            rmat = torch.einsum("xyab,xycda,xyedb->xyce",lmat,mps_top[:,j,:,:,:,:],mps_bottom[:,j,:,:,:,:]).contiguous().view(batch,10,mps_top.shape[3],mps_bottom.shape[3])
#        lmat = lmat/lmat.norm()
#        rmat = rmat/rmat.norm()
        x = torch.einsum("xyab,xyab->xy",lmat,rmat)
        print(x)
        print(torch.max(x),torch.min(x))
        print("x.norm=",x.norm())
        input("x")
#        x = x/x.norm()
        print("x.norm=",x.norm())
#        print("x:",x[0,:10])
#        print("before return, x.shape=",x.shape)
#        input("xyz")
        print(x)
    #    print(F.log_softmax(x*self.beta.item(), dim=1))
    #    return F.log_softmax(x*self.beta.item(), dim=1)
        return torch.log(torch.abs(x))

#    @weak_script_method
    def forward(self, x_input):
        tensors=torch.einsum("abcd,cdbeijkl->acdeijkl",x_input,self.tensors) # batch,l,l,10,left,down,right,up
        batch = tensors.shape[0]
        D=self.D
        L=self.L
#        factor = torch.zeros(batch
        mps_top = torch.zeros(batch,L,10,D,D,D,dtype=x_input.dtype,device=x_input.device) # batch,l,10,left,down,right
        mps_top = tensors[:,0,:,:,:,:,:,0]# batch,l,10,left,down,right
        mps_bottom = torch.zeros(batch,L,10,D,D,D,dtype=x_input.dtype,device=x_input.device) # batch,l,10,left,up,right
        mps_bottom = tensors[:,L-1,:,:,:,0,:,:]# batch,l,10,left,down,right
        for i in range(1,L>>1):
            mps_top_new = torch.zeros(batch,L,10,mps_top.shape[3]*D,D,mps_top.shape[5]*D,dtype=x_input.dtype,device=x_input.device) # batch,l,10,left,down,right
            mps_bottom_new = torch.zeros(batch,L,10,mps_bottom.shape[3]*D,D,mps_bottom.shape[5]*D,dtype=x_input.dtype,device=x_input.device) # batch,l,10,left,down,right
            for j in range(L):
                mps_top_new[:,j,:,:,:,:] = torch.einsum("xyabc,xyijkb->xyaijck",mps_top[:,j,:,:,:,:],tensors[:,i,j,:,:,:,:,:]).contiguous().view(batch,10,mps_top.shape[3]*D,D,mps_top.shape[5]*D) # x:batch y:10 
                mps_bottom_new[:,j,:,:,:,:] = torch.einsum("xyabc,xyibjk->xyiakjc",mps_bottom[:,j,:,:,:,:],tensors[:,L-1-i,j,:,:,:,:,:]).contiguous().view(batch,10,mps_bottom.shape[3]*D,D,mps_bottom.shape[5]*D) # x:batch y:10 
            oldshape=mps_top_new.shape
#            print("mps_top_new.norm=",mps_top_new.norm())
            mps_top_new = mps_top_new.view(mps_top_new.shape[0],-1)
            mymax = torch.max(torch.abs(mps_top_new),1,True).values
#            print("mymax=",mymax)
#            print(mps_top_new.shape,mymax.shape)
            mps_top_new = (mps_top_new/mymax).view(oldshape)
#            print("mps_top_new.norm=",mps_top_new.norm())
#            input("LLL")
            mps_top = mps_top_new


            oldshape=mps_bottom_new.shape
            mps_bottom_new = mps_bottom_new.view(mps_bottom_new.shape[0],-1)
            mymax = torch.max(torch.abs(mps_bottom_new),1,True).values
            mps_bottom_new = (mps_bottom_new/mymax).view(oldshape)
            mps_bottom = mps_bottom_new
        lmat = torch.einsum("xyab,xyac->xybc",mps_top[:,0,:,0,:,:],mps_bottom[:,0,:,0,:,:])
        rmat = torch.einsum("xyab,xycb->xyac",mps_top[:,0,:,:,:,0],mps_bottom[:,0,:,:,:,0])
#        print(lmat)
#        input("lmat")
        mymax = torch.max(torch.abs(lmat),1,True).values
        lmat = lmat/mymax
        mymax = torch.max(torch.abs(rmat),1,True).values
        rmat = rmat/mymax
        
        for j in range(1,L>>1):
            lmat = torch.einsum("xyab,xyacd,xybcf->xydf",lmat,mps_top[:,j,:,:,:,:],mps_bottom[:,j,:,:,:,:]).contiguous().view(batch,10,mps_top.shape[5],mps_bottom.shape[5])
            rmat = torch.einsum("xyab,xycda,xyedb->xyce",lmat,mps_top[:,j,:,:,:,:],mps_bottom[:,j,:,:,:,:]).contiguous().view(batch,10,mps_top.shape[3],mps_bottom.shape[3])
            mymax = torch.max(torch.abs(lmat),1,True).values
            lmat = lmat/mymax
            mymax = torch.max(torch.abs(rmat),1,True).values
            rmat = rmat/mymax
#            print(lmat)
#            input("***")
        x = torch.einsum("xyab,xyab->xy",lmat,rmat)

#        mymax = torch.max(torch.abs(x),1,True).values # normalize x will increase the generalization power, but decrease the expressive power
#        x = x/mymax

#        print(x)
#        input("x")
#        print(torch.log(torch.abs(x)))
    #    print(F.log_softmax(x*self.beta.item(), dim=1))
        return F.log_softmax(x*self.beta.item(), dim=1)
#        return F.log_softmax(x, dim=1)
#        return x
#        return torch.log(torch.abs(x))







class Net(nn.Module):
    def __init__(self,args):
        super(Net, self).__init__()
        self.n_channel_1=16
        self.n_channel_2=50
        padding_1=0
        padding_2=2
        self.fc_size=6
        self.conv1 = nn.Conv2d(1, self.n_channel_1, 5, 1,padding=padding_1)
        self.conv2 = nn.Conv2d(self.n_channel_1, self.n_channel_2, 5, 1,padding=padding_2)
        self.pepo = PEPO(self.fc_size,self.n_channel_2,3)

    def dropout(self,rate):
        self.filter = (torch.rand(self.pepo.tensors.shape,dtype=self.pepo.tensors.dtype,device=self.pepo.tensors.device)<rate)
        #self.filter        #self.pepo.tensors[self.filter].data = 0
        self.pepo.tensors[self.filter].data = self.pepo.tensors[self.filter].data-self.pepo.tensors[self.filter].data
        #self.pepo.tensors = self.pepo.tensors * self.filter

    def forward(self, x):
        x = F.relu(self.conv1(x))
#        print(x.shape)
        x = F.max_pool2d(x, 2, 2)
#        print(x.shape)
        x = F.relu(self.conv2(x))
#        print(x.shape)
        x = F.max_pool2d(x, 2, 2) ## after this step, the data has shape [ batch, channel, 6, 6]
#        print("input to pepo",x.shape)
        x = self.pepo(x)
#        print(x.shape)
        return F.log_softmax(x, dim=1)

